import numpy as np
import pytest

from nuclei.client.utils import serialize_jsonifyable_object


@pytest.mark.parametrize(
    "schema,expected",
    [
        (1, 1),
        (np.int8(1), int(1)),
        (np.int16(1), int(1)),
        (np.int32(1), int(1)),
        (np.int64(1), int(1)),
        (1.0, 1.0),
        (np.float32(1), float(1)),
        (np.float64(1), float(1)),
        ("str", "str"),
        (np.nan, None),
        (np.inf, None),
        ([1, 2, 3], [1, 2, 3]),
        ((1, 2, 3), [1, 2, 3]),
        ([1.0, 2.0, 3.0], [1.0, 2.0, 3.0]),
        ((1.0, 2.0, 3.0), [1.0, 2.0, 3.0]),
        (
            [1, np.int8(2), np.int16(3), np.int32(4), np.int64(5)],
            [1, 2, 3, 4, 5],
        ),
        (
            np.array([1, np.int8(2), np.int16(3), np.int32(4), np.int64(5)]),
            [1, 2, 3, 4, 5],
        ),
        (
            [1.0, np.float32(2), np.float64(3)],
            [1.0, 2.0, 3.0],
        ),
        (
            np.array([1.0, np.float32(2), np.float64(3)]),
            [1.0, 2.0, 3.0],
        ),
        (
            [True, np.bool_(True), False, np.bool_(False)],
            [True, True, False, False],
        ),
        (
            np.array([True, np.bool_(True), False, np.bool_(False)]),
            [True, True, False, False],
        ),
        (
            np.array([np.array([np.int16(1), 2.0]), np.array([np.nan, np.float32(4)])]),
            [[1.0, 2.0], [None, 4.0]],
        ),
        (
            {"str": "str", "list": np.array([np.nan, np.bool_(False), np.int32(3)])},
            {"str": "str", "list": [None, False, 3]},
        ),
    ],
)
def test_serialize(schema, expected):
    jsonifyable_object = serialize_jsonifyable_object(schema)
    assert jsonifyable_object == expected


def test_python_types_to_message_bool():
    # Test convert to jsonifyable object
    jsonifyable_object = serialize_jsonifyable_object(np.bool_(True))
    assert jsonifyable_object is True


def test_python_types_to_message_error():
    class Custom:
        def __init__(self):
            pass

    with pytest.raises(TypeError):
        serialize_jsonifyable_object(Custom())



import logging
import os

import jwt
import requests


def create_session() -> requests.Session:
    """
    Initialising session object to call the NUCLEI endpoints.
    Provide the user token from https://nuclei.cemsbv.io/#/personal-access-tokens to
    upon request or set as environment variable.

    Returns
    -------
    Session
    """

    # initialising session
    _session = requests.Session()

    # set bearer token
    _session.headers.update({"Authorization": f"Bearer {authenticate()}"})

    return _session


def authenticate() -> str:
    """
    Returns a validated JWT token from backend.

    Prompt the user for a user-token if it is not stored as an environmental variable
    "NUCLEI_TOKEN". A prompted user token will be stored as env-var after validation.

    Will throw an exception when authentication fails.
    """

    user_token = _get_valid_user_token()

    return user_token


def _get_valid_user_token() -> str:
    """
    Obtains the user-token from the env-vars or stdin and returns it.
    If the token is valid, it is stored as env-var "NUCLEI_TOKEN" as a side-effect.
    """

    # check if User Token is in environment variables
    if "NUCLEI_TOKEN" in os.environ:
        logging.info("user token found in environment")
        token = os.environ["NUCLEI_TOKEN"]
        _validate_user_token(token)

    # ask user for user token
    else:
        token = input(
            "Authentication is needed! Please provide your NUCLEI User Token. "
            "You can obtain your NUCLEI User Token on https://nuclei.cemsbv.io/#/personal-access-tokens."
        )

        _validate_user_token(token)
        os.environ["NUCLEI_TOKEN"] = token
        logging.info("user token set in environment")

    return token


def _validate_user_token(token: str) -> None:
    """Validate a JWT User-token by trying to decode it.

    Parameters
    ----------
    token: str
        The Nuclei User JWT.
    """

    try:
        _ = jwt.decode(
            token,
            algorithms=["HS256"],
            options={"verify_signature": False, "verify_exp": True},
        )
    except jwt.ExpiredSignatureError:
        raise jwt.ExpiredSignatureError(
            "Your NUCLEI User Token is expired. Please create a new token on "
            "https://nuclei.cemsbv.io/#/personal-access-tokens."
        )

    except jwt.InvalidTokenError as e:
        raise jwt.InvalidTokenError(
            "Your NUCLEI User Token is invalid. Please copy your user token from "
            "https://nuclei.cemsbv.io/#/personal-access-tokens. "
            f"The following error occurred: {e}"
        )



from __future__ import annotations

import base64
import time
from functools import lru_cache
from typing import Any, List, Literal, Optional, Union

import jwt

from nuclei import create_session

# try import serialize functions
try:
    from IPython.display import Image

    from nuclei.client import utils
except ImportError as e:
    raise ImportError(
        "Could not import one of dependencies [numpy, orjson, ipython]. "
        "You must install nuclei[client] in order to use NucleiClient \n"
        rf"Traceback: {e}"
    )

ROUTING = {
    "PileCore": {
        "v2": "https://crux-nuclei.com/api/pilecore/v2",
        "v3": "https://crux-nuclei.com/api/pilecore/v3",
        "latest": "https://crux-nuclei.com/api/pilecore/v3",
    },
    "VibraCore": {
        "v2": "https://crux-nuclei.com/api/vibracore/v2",
        "latest": "https://crux-nuclei.com/api/vibracore/v2",
    },
    "CPT Core": {
        "v1": "https://crux-nuclei.com/api/cptcore/v1",
        "latest": "https://crux-nuclei.com/api/cptcore/v1",
    },
    "ShallowCore": {
        "v1": "https://crux-nuclei.com/api/shallowcore/v1",
        "latest": "https://crux-nuclei.com/api/shallowcore/v1",
    },
}

DEFAULT_REQUEST_TIMEOUT = 10
MAX_RETRIES = 10


class NucleiClient:
    def __init__(self) -> None:
        """
        NUCLEI Client.

        This class allows the user to interact with our APIs. API documentation can be
        found on our platform > https://nuclei.cemsbv.io/#

        Attributes
        -----------
        session: Session
            requests.Session with authorisation set
        routing: dict
            Routing table to all available API's in the Nuclei landscape.
        timeout: int
            The connect timeout is the number of seconds. Default is 5 seconds
        """

        # initialize session
        self.session = create_session()

        # get routing table to application
        self.routing = ROUTING

        # set default timeout
        self.timeout = DEFAULT_REQUEST_TIMEOUT

    def get_url(self, app: str, version: str = "latest") -> str:
        """
        Get API's url

        Parameters
        ----------
        app: str
            Name of the API. Call `applications` to obtain a list with all applications.
        version: str, optional
            default is latest.
            API version used. Call `get_versions` to obtain a list with all version of a specific application.

        Returns
        -------
        url : str

        Raises
        -------
        TypeError:
            Wrong type for `app` or `version` argument
        ValueError:
            Wrong value for `app` or `version` argument
        """
        if not isinstance(version, str):
            raise TypeError(
                f"Expected positional argument `version` to be of type <class 'str'>, but got type: {type(version)}"
            )
        if version not in self.get_versions(app):
            raise ValueError(
                f"Application version not available, please select one of the following valid versions {self.get_versions(app)}"
            )
        return self.routing[app][version]

    @property
    def user_permissions(self) -> List[str | None]:
        """
        Provide the user permissions of your token.

        Returns
        -------
        out : list[str]
            Names of the API's
        """
        return jwt.decode(
            self.session.headers["Authorization"].split(" ")[1],  # type: ignore
            algorithms=["HS256"],
            options={"verify_signature": False, "verify_exp": False},
        ).get("permissions", [])

    @property
    def applications(self) -> List[str]:
        """
        Provide available API's in the Nuclei landscape.

        Returns
        -------
        out : list[str]
            Names of the API's
        """
        return list(self.routing.keys())

    def get_versions(self, app: str) -> List[str]:
        """
        Provide available API's versions in the Nuclei landscape.

        Parameters
        ----------
        app: str
            Name of the API. Call `applications` to obtain a list with all applications.

        Returns
        -------
        out : list[str]
            Versions of the API.

        Raises
        -------
        TypeError:
            Wrong type for `app` argument
        ValueError:
            Wrong value for `app` argument
        """
        if not isinstance(app, str):
            raise TypeError(
                f"Expected positional argument `app` to be of type <class 'str'>, but got type: {type(app)}"
            )
        if app not in self.applications:
            raise ValueError(
                f"Application not available, please select one of the following valid applications {self.applications}"
            )
        return list(self.routing[app].keys())

    @lru_cache(16)
    def _get_app_specification(self, app: str, version: str = "latest") -> dict:
        """
        Private methode to get the JSON schema of the API documentation.

        Parameters
        ----------
        app: str
            Name of the API. Call `applications` to obtain a list with all applications.
        version: str, optional
            default is latest.
            API version used. Call `get_versions` to obtain a list with all version of a specific application.

        Returns
        -------
        dict

        Raises
        -------
        ConnectionError:
            Application not available
        TypeError:
            Wrong type for `app` or `version` argument
        ValueError:
            Wrong value for `app` or `version` argument
        """
        response = self.session.get(
            self.get_url(app, version) + "/openapi.json", timeout=self.timeout
        )
        if response.status_code != 200:
            raise ConnectionError(
                "Unfortunately the application you are trying to reach is unavailable (status code: "
                f"{response.status_code}). Please check you connection. If the problem persists contact "
                "CEMS at info@cemsbv.nl"
            )
        return response.json()

    def get_application_version(self, app: str, version: str = "latest") -> str:
        """
        Provide the semanctic version of the API in the Nuclei landscape.

        Parameters
        ----------
        app: str
            Name of the API. Call `applications` to obtain a list with all applications.
        version: str, optional
            default is latest.
            API version used. Call `get_versions` to obtain a list with all version of a specific application.

        Returns
        -------
        out : str
            Semantic Version of the API

        Raises
        -------
        ConnectionError:
            Application not available
        TypeError:
            Wrong type for `app` or `version` argument
        ValueError:
            Wrong value for `app` or `version` argument
        """
        return self._get_app_specification(app, version)["info"]["version"]

    def get_endpoints(self, app: str, version: str = "latest") -> List[str]:
        """
        Get available endpoints of single API.

        Parameters
        ----------
        app: str
            Name of the API. Call `applications` to obtain a list with all applications.
        version: str, optional
            default is latest
            API version used. Call `get_versions` to obtain a list with all version of a specific application.

        Returns
        -------
        endpoints : list[str]
            Endpoint urls.

        Raises
        -------
        ConnectionError:
            Application not available
        TypeError:
            Wrong type for `app` or `version` argument
        ValueError:
            Wrong value for `app` or `version` argument
        """
        return list(self._get_app_specification(app, version)["paths"].keys())

    def get_endpoint_type(
        self, app: str, endpoint: str, version: str = "latest"
    ) -> List[str]:
        """
        Get a list of HTTP methods used for this endpoint.

        Parameters
        ----------
        app: str
            Name of the API. Call `applications` to obtain a list with all applications.
        endpoint
            url of the endpoint.
        version: str, optional
            default is latest.
            API version used. Call `get_versions` to obtain a list with all version of a specific application.

        Returns
        -------
        methods: List[str]
            List of HTTP methods used for endpoint

        Raises
        -------
        ConnectionError:
            Application not available
        TypeError:
            Wrong type for an argument
        ValueError:
            Wrong value for an argument
        """
        if not isinstance(endpoint, str):
            raise TypeError(
                f"Expected positional argument `endpoint` to be of type <class 'str'>, but got type: {type(endpoint)}"
            )

        if endpoint in self.get_endpoints(app, version):
            return list(
                self._get_app_specification(app, version)["paths"][endpoint].keys()
            )
        raise ValueError(
            f"Endpoint name not valid, please select on of the following valid endpoints {self.get_endpoints(app, version)}"
        )

    def call_endpoint(
        self,
        app: str,
        endpoint: str,
        methode: Literal["auto", "get", "post"] = "auto",
        version: str = "latest",
        schema: Optional[Union[dict, str]] = None,
        return_response: bool = False,
    ) -> Any:
        """
        Calls an API in the nuclei landscape.

        Note
        --------
        The provided `schema` is serialized by the `utils.serialize_jsonifyable_object`. For get request the
        payload is passed to the `params` parameter of `request.get` function. For post request the payload
        is passed to the `json`  parameter of `request.post` function. This has an effect on the request content type.

        Parameters
        ----------
        app: str
            Name of the API. Call `applications` to obtain a list with all applications.
        endpoint: str
            Name of the API's endpoint. call `get_endpoints` to obtain a list with all applications for a given API.
        methode: str
            default is auto
            HTTP methode used to call endpoint. When auto methode is selected the HTTP methode is
            obtained from the openapi docs. Please note that this is the first one. call `get_endpoint_type`
            to obtain list with all methods related to the endpoint.
        version: str, optional
            default is latest.
            API version used. Call `get_versions` to obtain a list with all versions of a specific application.
        schema: dict or json-string, optional
            Default is None
            The parameter schema for the API. Take a look at the API documentation.
        return_response: bool, optional
            Default is False
            Return the requests response instead of the parsed data object.

        Returns
        -------
        json : dict
            json response
        text : str
            text response
        content : bytes
            content response
        out : Response
            requests response object
        figure: Image
             IPython display Image object

        Raises
        -------
        RuntimeError:
            Thrown when response is between 400 and 600 to see if
            there was a client error or a server error.
        NotImplementedError:
            HTTP methode not get or post request
        ConnectionError:
            Application not available
        TypeError:
            Wrong type for an argument
        ValueError:
            Wrong value for an argument
        """
        if not isinstance(endpoint, str):
            raise TypeError(
                f"Expected positional argument `endpoint` to be of type <class 'str'>, but got type: {type(endpoint)}"
            )

        if endpoint not in self.get_endpoints(app, version):
            raise ValueError(
                f"Endpoint name not valid, please select on of the following valid endpoints {self.get_endpoints(app, version)}"
            )

        if not isinstance(methode, str):
            raise TypeError(
                f"Expected keyword-argument `methode` to be of type <class 'str'>, but got type: {type(methode)}"
            )
        if methode not in ["auto", "get", "post"]:
            raise ValueError(
                f'Expected value of keyword-argument `methode` to be one of ["auto", "get", "post"] , but got: {methode}'
            )

        if not (schema is None or isinstance(schema, (str, dict))):
            raise TypeError(
                f"Expected keyword-argument `schema` to be of type <class 'dict'> or <class 'str'>, but got type: {type(schema)}"
            )

        if not isinstance(return_response, bool):
            raise TypeError(
                f"Expected keyword-argument `return_response` to be of type <class 'bool'>, but got type: {type(return_response)}"
            )

        if methode == "auto":
            t = self.get_endpoint_type(app=app, version=version, endpoint=endpoint)[0]
        else:
            t = methode

        if isinstance(schema, str):
            schema = utils.to_json(schema)

        if t.lower() not in ["get", "post"]:
            raise NotImplementedError(
                "Not a valid HTTP request methode. Only GET or POST requests are supported. "
                f"Use the session attribute to get full control of your request. Provided methode: {t}"
            )

        execution_count = 0

        # We enter a retry-loop for the execution of the main call, with a MAX_RETRIES
        # being the maximum number of retires
        while execution_count < MAX_RETRIES:
            execution_count += 1

            if t.lower() == "get":
                response = self.session.get(
                    self.get_url(app=app, version=version) + endpoint,
                    params=utils.serialize_jsonifyable_object(schema),
                    timeout=self.timeout,
                )
            elif t.lower() == "post":
                response = self.session.post(
                    self.get_url(app=app, version=version) + endpoint,
                    json=utils.serialize_jsonifyable_object(schema),
                    timeout=self.timeout,
                )

            # If the response contains one of the following statuses, retry the request.
            #   429 Too Many Requests
            #   502 Bad Gateway
            #   503 Service Unavailable
            #   504 Gateway Timeout
            status_codes_that_trigger_retry = [429, 502, 503, 504]
            if response.status_code in status_codes_that_trigger_retry:
                # If the call failed, and retry is a viable option, we sleep for a
                # little while and execute another call
                time.sleep(1)
                continue

            # We break the retry loop by default
            break

        if return_response:
            return response

        if not response.ok:
            raise RuntimeError(
                "An error was thrown during your request. Please take a look at the response object for "
                "more information. You can get the response object by setting the `return_response` attribute"
                "to True. \n"
                f"Request URL: {response.url} \n"
                f"Status code: {response.status_code} \n"
                f"Message:     {response.content[:100]!r}"
            )

        content_type = response.headers["Content-Type"]
        if content_type == "image/png;base64":
            return Image(base64.b64decode(response.text))
        elif content_type == "image/png":
            return Image(response.content)
        elif content_type.endswith("json"):
            return response.json()
        elif content_type.startswith("text/"):
            return response.text
        return response.content



import io
import json
import os

import pytest
import requests
from IPython.display import Image

from nuclei.client.main import NucleiClient

from .conftest import mock_valid_jwt


def test_client_env_vars(user_token_envvar):
    """Tests if a NucleiClient can be instantiated with correct token in env-var."""

    client = NucleiClient()

    assert isinstance(client.session, requests.Session)


def test_client_stdin(monkeypatch):
    """Tests if a NucleiClient can be instantiated after passing a token through stdin"""

    monkeypatch.setattr(os, "environ", {})
    monkeypatch.setattr("sys.stdin", io.StringIO(mock_valid_jwt(key="user_secret")))

    client = NucleiClient()

    assert isinstance(client.session, requests.Session)
    assert "NUCLEI_TOKEN" in os.environ.keys()


@pytest.mark.parametrize("app", ["PileCore", "VibraCore", "CPT Core"])
def test_applications_to_endpoints(app, user_token_envvar):
    """Tests correct responses for a single application of four endpoints:
    - get_applications: Is the appname in the list?,
    - get_url: Is an url returned for the appname?,
    - get_endpoints: Is a list of endpoints returned for the response of get_url?
    - get_endpoint_type: Does get_endpoint_type return a correct response for each returned endpoint?
    """

    client = NucleiClient()

    # Test get_applications
    apps = client.applications

    assert app in apps

    # Test get_url
    assert isinstance(client.get_url(app, version="latest"), str)

    # Test get_endpoints. Assumes pulic openapi docs for all APIs.
    endpoints = client.get_endpoints(app, version="latest")

    for endpoint in endpoints:
        assert isinstance(endpoint, str)
        assert endpoint[0] == r"/"

        endpoint_type = client.get_endpoint_type(
            app, version="latest", endpoint=endpoint
        )[0]

        assert endpoint_type in ("get", "post")


@pytest.mark.parametrize("app", ["pilecore", "vibracore", "CPTCore"])
def test_get_wrong_url(user_token_envvar, app):
    "Tests if a ValueError is raised after passing an erroneous appname"

    client = NucleiClient()

    with pytest.raises(ValueError) as err:
        client.get_url(app, version="latest")

    assert str(err.value).startswith(
        "Application not available, please select one of the following valid applications"
    )


@pytest.mark.parametrize("version", ["v0"])
def test_get_wrong_version(user_token_envvar, version):
    "Tests if a ValueError is raised after passing an erroneous app version"

    client = NucleiClient()

    with pytest.raises(ValueError) as err:
        client.get_url("PileCore", version=version)

    assert str(err.value).startswith(
        "Application version not available, please select one of the following valid versions"
    )


def test_get_userclaims(user_token_envvar):
    """Tests if the get_user_claims endpoint returns the correct claims for a free user"""

    client = NucleiClient()

    assert client.user_permissions == ["read:users"]


def test_wrong_endpoint_for_type(user_token_envvar):
    """Tests if a ValueError is raised after calling get_endpoint_type with an
    non-existing endpoint"""

    client = NucleiClient()

    with pytest.raises(ValueError) as err:
        client.get_endpoint_type(
            "PileCore", version="latest", endpoint="/some/invalid/endpoint"
        )

    assert str(err.value).startswith(
        "Endpoint name not valid, please select on of the following valid endpoints"
    )


def test_call_endpoint_post_returns_json(
    user_token_envvar,
    get_app_specification_post,
    session_send_post_returns_json,
):
    """Test `call_endpoint`: Post endpoint usage with dictionary schema"""

    client = NucleiClient()

    response = client.call_endpoint(
        app="PileCore", endpoint="/MockPostEndpoint", schema={"somekey": "somevalue"}
    )

    assert isinstance(response, dict)
    assert response["message"] == "Test Confirmed"


def test_call_endpoint_get_returns_json(
    user_token_envvar,
    get_app_specification_get,
    session_send_get_returns_json,
):
    """Test `call_endpoint`: Get endpoint usage with dictionary schema"""

    client = NucleiClient()

    response = client.call_endpoint(
        app="PileCore", endpoint="/MockGetEndpoint", schema={"somekey": "somevalue"}
    )

    assert isinstance(response, dict)
    assert response["message"] == "Test Confirmed"


def test_call_endpoint_string_schema_returns_json(
    user_token_envvar,
    get_app_specification_post,
    session_send_post_returns_json,
):
    """Test `call_endpoint`: post endpoint usage with string schema"""

    client = NucleiClient()

    response = client.call_endpoint(
        app="PileCore",
        endpoint="/MockPostEndpoint",
        schema=json.dumps({"somekey": "somevalue"}),
    )

    assert isinstance(response, dict)
    assert response["message"] == "Test Confirmed"


def test_call_endpoint_return_response(
    user_token_envvar,
    get_app_specification_post,
    session_send_post_returns_json,
):
    """Test `call_endpoint`: usage with return_response flag"""

    client = NucleiClient()

    response = client.call_endpoint(
        app="PileCore",
        endpoint="/MockPostEndpoint",
        schema={"somekey": "somevalue"},
        return_response=True,
    )

    assert isinstance(response, requests.Response)
    assert response.status_code == 200
    assert response.json()["message"] == "Test Confirmed"


def test_call_endpoint_invalid_method(
    user_token_envvar,
    get_app_specification_invalid_method,
):
    """Test `call_endpoint`: Post endpoint usage with invalid method"""

    client = NucleiClient()

    with pytest.raises(NotImplementedError) as err:
        client.call_endpoint(app="PileCore", endpoint="/MockWeirdEndpoint")

    assert (
        str(err.value)
        == "Not a valid HTTP request methode. Only GET or POST requests are supported. "
        "Use the session attribute to get full control of your request. Provided methode: weird"
    )


def test_call_endpoint_post_returns_png(
    user_token_envvar,
    get_app_specification_post,
    session_send_post_returns_png,
):
    """Test `call_endpoint`: Post endpoint usage with dictionary schema.
    Server response is a byte64 encoded png, which nuclei should decode."""

    client = NucleiClient()

    response = client.call_endpoint(
        app="PileCore", endpoint="/MockPostEndpoint", schema={"somekey": "somevalue"}
    )

    assert isinstance(response, Image)


def test_call_endpoint_post_returns_b64_png(
    user_token_envvar,
    get_app_specification_post,
    session_send_post_returns_b64_png,
):
    """Test `call_endpoint`: Post endpoint usage with dictionary schema
    Server response is a png, which nuclei should pass"""

    client = NucleiClient()

    response = client.call_endpoint(
        app="PileCore", endpoint="/MockPostEndpoint", schema={"somekey": "somevalue"}
    )

    assert isinstance(response, Image)


def test_call_endpoint_post_returns_json_with_parsing_error(
    user_token_envvar,
    get_app_specification_post,
    session_send_post_returns_json,
    mock_serialize_json_bytes_parsing_error,
):
    client = NucleiClient()

    with pytest.raises(Exception):
        client.call_endpoint(
            app="PileCore",
            endpoint="/MockPostEndpoint",
            schema={"somekey": "somevalue"},
        )


def test_call_endpoint_post_returns_text(
    user_token_envvar,
    get_app_specification_post,
    session_send_post_returns_text,
):
    client = NucleiClient()

    response = client.call_endpoint(
        app="PileCore", endpoint="/MockPostEndpoint", schema={"somekey": "somevalue"}
    )

    assert isinstance(response, str)
    assert response == "Some text"


def test_call_endpoint_version(
    user_token_envvar,
    get_app_specification_version,
):
    """Test `get_application_version`"""

    client = NucleiClient()

    assert (
        client.get_application_version(app="PileCore", version="latest")
        == "0.1.0-beta.1"
    )
